#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>		// printk()
#include <linux/slab.h>			// kmalloc()
#include <linux/fs.h>			// 
#include <linux/errno.h>		// error code
#include <linux/types.h>		// size_t
#include <linux/proc_fs.h>
#include <linux/fcntl.h>		// O_ACCMODE
#include <linux/cdev.h>

#include <asm/system.h>			// cli(), *_flags
#include <asm/uaccess.h>		// copy_*_user()

#include "template.h"

int template_major		= TEMPLATE_MAJOR;
int template_minor		= 0;
int template_nr_devs	= 0;
char global_data[100];

module_param( template_major, int, S_IRUGO );
module_param( template_minor, int, S_IRUGO );

MODULE_AUTHOR( "Choi, Jae-woo" );
MODULE_LICENSE( "Dual BSD/GPL" );

int template_open( struct inode *inode, struct file *filp )
{
	return 0;
}

int template_release( struct inode *inode, struct file *filp )
{
	return 0;
}

ssize_t template_read( struct file *filp, char __user *buf, size_t count, loff_t *f_pos )
{
	ssize_t retval = 0;


	// down_interruptible();

	// .....

	if ( copy_from_user( global_data, buf, count ) )
	{
		retval = -EFAULT;
		goto out;
	}

	*f_pos += count;
	retval = count;

out:
	// up();

	return retval;
}

ssize_t template_write( struct file *filp, const char __user *buf, size_t count, loff_t *f_pos )
{
	ssize_t retval = 0;

	// down_interruptible();

	// .....

	if ( copy_from_user( global_data, buf, count ) )
	{
		retval = -EFAULT;
		goto out;
	}

	*f_pos += count;
	retval = count;

out:
	// up( );

	return retval;
}

int template_ioctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg )
{
	int err = 0;
	int retval = 0;

	if ( _IOC_TYPE( cmd ) != TEMPLATE_IOC_MAGIC )
		return -ENOTTY;

	if ( _IOC_NR( cmd ) > TEMPLATE_IOC_MAXNR )
		return -ENOTTY;

	if ( _IOC_DIR( cmd ) & _IOC_READ )
		err = ! access_ok( VERIFY_WRITE, (void __user *) arg, _IOC_SIZE( cmd ) );

	else if ( _IOC_DIR( cmd ) & _IOC_WRITE )
		err = ! access_ok( VERIFY_READ, (void __user *) arg, _IOC_SIZE( cmd ) );

	if ( err )
		return -EFAULT;

	switch ( cmd )
	{
		case TEMPLATE_IOCRESET:
			break;

		default:
			return -ENOTTY;
			
	}

	return retval;
}

loff_t template_llseek( struct file *filp, loff_t off, int whence )
{
	loff_t newpos;

	switch ( whence )
	{
		case 0 :		// SEEK_SET
			newpos = off;
			break;

		case 1 :		// SEEK_CUR
			newpos = filp->f_pos + off;
			break;

		case 2 :		// SEEK_END
			newpos = sizeof( global_data ) + off;
			break;

		default :
			return -EINVAL;
	}

	if ( newpos < 0 )
		return -EINVAL;

	filp->f_pos = newpos;

	return newpos;
}


struct file_operations template_fops =
{
	.owner			= THIS_MODULE,
	.llseek			= template_llseek,
	.read			= template_read,
	.write			= template_write,
	.ioctl			= template_ioctl,
	.open			= template_open,
	.release		= template_release,
};

static void template_cleanup_module( void )
{
	dev_t devno = MKDEV( template_major, template_minor );


//	template_remove_proc( );

	unregister_chrdev_region( devno, template_nr_devs );

}

static int template_init_module( void )
{
	int result;
	dev_t dev = 0;

	result = alloc_chrdev_region( &dev, template_minor, template_nr_devs, "template" );
	template_major = MAJOR( dev );

	if ( result < 0 )
	{
		printk( KERN_WARNING "template: can't get major %d\n", template_major );

		return result;
	}

	dev = MKDEV( template_major, template_minor + template_nr_devs );

//	template_create_proc( );

	return 0;
}


module_init( template_init_module );
module_exit( template_cleanup_module );


